# Generated by Django 4.1.3 on 2023-02-07 07:01

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("alastria_auth", "0002_apisession_request_session_id_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="apisession",
            name="extra_info",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="apisession",
            name="expiration_date",
            field=models.DateTimeField(
                default=datetime.datetime(
                    2023, 2, 10, 7, 1, 36, 872311, tzinfo=datetime.timezone.utc
                ),
                verbose_name="Expirate date",
            ),
        ),
        migrations.AlterField(
            model_name="requestsession",
            name="expiration_date",
            field=models.DateTimeField(
                default=datetime.datetime(
                    2023, 2, 7, 7, 6, 36, 872311, tzinfo=datetime.timezone.utc
                ),
                verbose_name="Expirate date",
            ),
        ),
    ]



# Generated by Django 4.1.3 on 2023-02-07 09:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("alastria_auth", "0003_apisession_extra_info_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="apisession",
            name="expiration_date",
            field=models.DateTimeField(verbose_name="Expirate date"),
        ),
        migrations.AlterField(
            model_name="apisession",
            name="token",
            field=models.CharField(max_length=1255),
        ),
        migrations.AlterField(
            model_name="requestsession",
            name="expiration_date",
            field=models.DateTimeField(verbose_name="Expirate date"),
        ),
    ]



from django.contrib import admin
from .models import RequestSession, ApiSession
import waffle
from .enums import AlastriaAuthSwitches


class RequestSessionAdmin(admin.ModelAdmin):
    model = RequestSession

    def get_model_perms(self, request):

        if not waffle.switch_is_active(AlastriaAuthSwitches.auth.value):
            return {}

        return super(RequestSessionAdmin, self).get_model_perms(request)


class ApiSessionAdmin(admin.ModelAdmin):
    model = ApiSession

    def get_model_perms(self, request):

        if not waffle.switch_is_active(AlastriaAuthSwitches.auth.value):
            return {}

        return super(ApiSessionAdmin, self).get_model_perms(request)


admin.site.register(RequestSession, RequestSessionAdmin)
admin.site.register(ApiSession, ApiSessionAdmin)



from django.conf import settings
from .models import RequestSession
from abc import ABC, abstractmethod
from typing import Any


class AAlastriaAuthService(ABC):
    @staticmethod
    @abstractmethod
    def validate_permission(token: str) -> None:
        ...

    @staticmethod
    @abstractmethod
    def get_did(address: str) -> str:
        ...

    @staticmethod
    @abstractmethod
    def validate_signature(token: str) -> Any:
        ...

    @staticmethod
    @abstractmethod
    def public_key_to_address(public_key: str) -> str:
        ...

    @staticmethod
    @abstractmethod
    def get_request_session_serializer() -> RequestSession:
        ...

    @staticmethod
    @abstractmethod
    def validate_did(did, response):
        ...

    @staticmethod
    @abstractmethod
    def validate_request_session(rs: RequestSession, format_did: str, response):
        ...



from rest_framework import serializers
from .models import RequestSession, ApiSession


class RequestSessionSerializer(serializers.ModelSerializer):
    class Meta:
        model = RequestSession
        fields = (
            "pk",
            "request_date",
            "expiration_date",
            "requested_by",
            "qr_code",
            "confirm_url",
        )


class RequestSerializer(serializers.ModelSerializer):
    class Meta:
        model = ApiSession
        fields = (
            "token",
            "expiration_date",
            "is_active",
        )


class SignatureSerializer(serializers.Serializer):
    token = serializers.CharField(max_length=1500)


class ConfirmSerializer(serializers.Serializer):
    request_session_id = serializers.IntegerField()


class ConfirmResponseSerializer(serializers.Serializer):
    api_session_token = serializers.CharField(max_length=255)


class SuccessResponseSerializer(serializers.Serializer):
    response = serializers.CharField(max_length=2)



from django.db import models
from datetime import timedelta
from django.utils.timezone import now


class RequestSession(models.Model):
    request_date = models.DateTimeField()
    expiration_date = models.DateTimeField(verbose_name="Expirate date")
    requested_by = models.CharField(max_length=255)
    qr_code = models.ImageField()
    confirm_url = models.CharField(max_length=255)
    completed = models.BooleanField(default=False)


class ApiSession(models.Model):

    token = models.CharField(max_length=1255)
    expiration_date = models.DateTimeField(verbose_name="Expirate date")
    is_active = models.BooleanField(default=True)
    request_session_id = models.IntegerField()
    extra_info = models.JSONField(blank=True, null=True)



from django.shortcuts import render
from rest_framework.viewsets import ViewSet
from rest_framework.permissions import AllowAny
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from .serializers import (
    RequestSessionSerializer,
    SignatureSerializer,
    ConfirmResponseSerializer,
    SuccessResponseSerializer,
    ConfirmSerializer,
)
from rest_framework.decorators import action
from .models import RequestSession
from rest_framework.response import Response
from .service import AlastriaAuthService
from .models import ApiSession
from .decorators import requires_alastria_auth
from waffle.decorators import waffle_switch
from .enums import AlastriaAuthSwitches


class AlastriaAuthView(ViewSet):
    permission_classes = (AllowAny,)

    @swagger_auto_schema(
        method="get",
        manual_parameters=[],
        responses={200: openapi.Response("", RequestSessionSerializer)},
    )
    @waffle_switch(AlastriaAuthSwitches.auth.value)
    @action(detail=False, methods=["get"])
    def auth(self, request):
        rs: RequestSession = AlastriaAuthService.get_request_session_serializer()
        rss: RequestSessionSerializer = RequestSessionSerializer(rs)
        return Response(rss.data)

    @swagger_auto_schema(
        operation_description="""
        Token field is a jwt signed by the subjet and signature alg should be ES256K.
        Example:
        {
            "header": {
                "alg": "ES256K",
                "typ": "JWT",
                "jwk": "0x03fdd57adec3d438ea237fe46b33ee1e016eda6b585c3e27ea66686c2ea5358479"
            },
            "payload": {
                "@context": ["https://alastria.github.io/identity/artifacts/v"],
                "iss": "did:ala:quor:redT:dc4eeee191cf4b1adeee4e91b25dec6dd8ec79c4",
                "pku": "",
                "iat": 1585128712,
                "nbf": 1585139712,
                "alastriaToken": "",
                "jti": "ze298y42sba"
            }
        }
        """,
        method="post",
        request_body=SignatureSerializer,
        responses={200: openapi.Response("", SuccessResponseSerializer)},
    )
    @waffle_switch(AlastriaAuthSwitches.auth.value)
    @action(detail=False, methods=["post"])
    def signature(self, request):
        token: str = request.data["token"]
        return AlastriaAuthService.validate_signature(token)

    @swagger_auto_schema(
        method="post",
        request_body=ConfirmSerializer,
        responses={200: openapi.Response("", ConfirmResponseSerializer)},
    )
    @waffle_switch(AlastriaAuthSwitches.auth.value)
    @action(detail=False, methods=["post"])
    def confirm(self, request):
        request_session_id: str = request.data["request_session_id"]
        api_session: ApiSession = ApiSession.objects.get(
            request_session_id=request_session_id
        )
        return Response({"response": api_session.token})

    @swagger_auto_schema(
        method="post", responses={200: openapi.Response("", SuccessResponseSerializer)}
    )
    @requires_alastria_auth
    @waffle_switch(AlastriaAuthSwitches.auth.value)
    @action(detail=False, methods=["post"])
    def log_out(self, request):
        return Response({"response": "OK"})



from django.core.exceptions import PermissionDenied
from functools import wraps
from .service import AlastriaAuthService
import waffle
from .enums import AlastriaAuthSwitches


def requires_alastria_auth(view):
    @wraps(view)
    def _view(request, *args, **kwargs):
        if not waffle.switch_is_active(AlastriaAuthSwitches.auth.value):
            return view(request, *args, **kwargs)
        if "HTTP_ALASTRIA_TOKEN" not in request.request.META.keys():  # Alastria-Token
            raise PermissionDenied
        else:
            AlastriaAuthService.validate_permission(
                request.request.META["HTTP_ALASTRIA_TOKEN"]
            )
        return view(request, *args, **kwargs)

    return _view



# Generated by Django 4.1.3 on 2023-01-13 08:23

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ApiSession',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.CharField(max_length=255)),
                ('expiration_date', models.DateTimeField(default=datetime.datetime(2023, 1, 16, 8, 23, 21, 999940, tzinfo=datetime.timezone.utc), verbose_name='Expirate date')),
                ('is_active', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='RequestSession',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('request_date', models.DateTimeField()),
                ('expiration_date', models.DateTimeField(default=datetime.datetime(2023, 1, 13, 8, 28, 21, 999126, tzinfo=datetime.timezone.utc), verbose_name='Expirate date')),
                ('requested_by', models.CharField(max_length=255)),
                ('qr_code', models.ImageField(upload_to='')),
                ('confirm_url', models.CharField(max_length=255)),
            ],
        ),
    ]



# Generated by Django 4.1.3 on 2023-01-16 07:19

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('alastria_auth', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='apisession',
            name='request_session_id',
            field=models.IntegerField(default=1),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='requestsession',
            name='completed',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='apisession',
            name='expiration_date',
            field=models.DateTimeField(default=datetime.datetime(2023, 1, 19, 7, 19, 45, 895576, tzinfo=datetime.timezone.utc), verbose_name='Expirate date'),
        ),
        migrations.AlterField(
            model_name='requestsession',
            name='expiration_date',
            field=models.DateTimeField(default=datetime.datetime(2023, 1, 16, 7, 24, 45, 894997, tzinfo=datetime.timezone.utc), verbose_name='Expirate date'),
        ),
    ]



from django.conf import settings
from uuid import uuid4
from .models import ApiSession, RequestSession
from django.utils.timezone import now
import jwt
from PIL import Image, ImageDraw
from django.core.files import File
from io import BytesIO
import qrcode
from .abstractions import AAlastriaAuthService
from web3 import Web3
from eth_keys import keys
from eth_utils import decode_hex
from alastria_identity.services import TokenService
from django.http.response import HttpResponseBadRequest
from rest_framework.response import Response
from alastria_service_client.client import AClient, Client
from typing import Any
from django.core.exceptions import PermissionDenied
from alastria_service_client.validators import (
    NetworkValidator,
    OnlyNetworkValidator,
    Address,
)
from datetime import timedelta
from alastria_identity.types import JwtToken, AlastriaToken
from network_service_client.client import (
    AClient as ANetworkClient,
    Client as NetworkClient,
    Network as NetworkDTO,
    NetworksNames,
)
from network_service_client.enums import ContractsNames
from datetime import datetime, timedelta


class AlastriaAuthService(AAlastriaAuthService):
    @staticmethod
    def public_key_to_address(public_key: str) -> str:
        return Web3.toChecksumAddress(
            keys.PublicKey(decode_hex(public_key[2:])).to_address()
        )

    @staticmethod
    def get_did(address: str) -> str:
        alastria_service_client: AClient = Client(
            service_host=settings.ALASTRIA_SERVICE_HOST
        )
        network_data: NetworkDTO = NetworkClient(
            service_host=settings.NETWORK_SERVICE_HOST
        ).get_network_by_name(NetworksNames.AlastriaDefaultName)
        network_body = NetworkValidator(
            provider=network_data.node["path"],
            identity_manager_contract_address=list(
                filter(
                    lambda contract: contract["name"]
                    == ContractsNames.AlastriaIdentityManager.value,
                    network_data.contracts,
                )
            )[0]["address"],
            identity_manager_contract_abi=list(
                filter(
                    lambda contract: contract["name"]
                    == ContractsNames.AlastriaIdentityManager.value,
                    network_data.contracts,
                )
            )[0]["abi"],
            public_key_registry_contract_address="",
            public_key_registry_contract_abi="",
            credential_registry_contract_abi="",
            credential_registry_contract_address="",
            chainId=network_data.chain_id,
        )
        return alastria_service_client.identity_keys(
            address=Address(address), body=OnlyNetworkValidator(network=network_body)
        ).response

    @staticmethod
    def validate_permission(token: str) -> None:
        token_service = TokenService(settings.ISSUER_PRIVATE_KEY)
        decoded = token_service.decode_jwt(token)
        session_id: str = decoded.get("payload").get("session_id")

        aps: ApiSession = ApiSession.objects.filter(pk=session_id).first()

        if not aps:
            return PermissionDenied

        if aps.expiration_date < now():
            raise PermissionDenied

        if not token_service.verify_jwt(token, settings.ISSUER_PUBLIC_KEY):
            raise PermissionDenied

    @staticmethod
    def get_request_session_serializer() -> RequestSession:
        rs: RequestSession = RequestSession(
            request_date=now(),
            expiration_date=now() + timedelta(minutes=5),
            requested_by="",
            confirm_url="/alastria/alastria-auth/confirm",
        )
        rs.save()
        network_data: NetworkDTO = NetworkClient(
            service_host=settings.NETWORK_SERVICE_HOST
        ).get_network_by_name(NetworksNames.AlastriaDefaultName)
        at: dict = AlastriaToken(
            network_data.did_prefix
            + AlastriaAuthService.get_did(settings.ISSUER_ADDRESS)[2:],
            network_data.node["path"],
            "/alastria/alastria-auth/confirm",
            settings.ALASTRIA_T_NETWORK_ID,
            1,
        ).build_jwt()
        at["payload"].update(
            {
                "signature_url": "/alastria/alastria-auth/signature",
                "request_session_id": rs.pk,
            }
        )
        token_service = TokenService(settings.ISSUER_PRIVATE_KEY)
        signed = token_service.sign_jwt(JwtToken(**at))
        qr_image = qrcode.make(signed)
        qr_offset = Image.new("RGB", (910, 910), "white")
        ImageDraw.Draw(qr_offset)
        qr_offset.paste(qr_image)
        file_name = f"{rs.requested_by}-{uuid4()}qr.png"
        stream = BytesIO()
        qr_offset.save(stream, "PNG")
        rs.qr_code.save(file_name, File(stream), save=False)
        qr_offset.close()
        rs.save()
        return rs

    @staticmethod
    def validate_did(did, response):
        if did and did[:4] != "0x00":
            return response
        else:
            return HttpResponseBadRequest("Not an alastria subject.")

    @staticmethod
    def validate_request_session(rs: RequestSession, format_did: str, response):
        if (
            rs
            and rs.requested_by == format_did
            and rs.expiration_date > now()
            and not rs.completed
        ):
            token_service = TokenService(settings.ISSUER_PRIVATE_KEY)

            rs.completed = True
            rs.save()
            aps: ApiSession = ApiSession(
                request_session_id=rs.pk, 
                expiration_date=datetime.now() + timedelta(days=3)
            )
            aps.save()
            header = {"alg": "ES256K", "typ": "JWT"}
            payload = {
                "expire_date": str(aps.expiration_date),
                "session_id": aps.pk,
                "request_session_id": aps.request_session_id,
                "extra_info": aps.extra_info,
            }
            signed = token_service.sign_jwt(JwtToken(header=header, payload=payload))
            aps.token = signed
            aps.save()
            return response
        else:
            return HttpResponseBadRequest("The request session is not valid.")

    @staticmethod
    def validate_signature(token: str) -> Any:
        token_service = TokenService(settings.ISSUER_PRIVATE_KEY)
        decoded = token_service.decode_jwt(token)
        alastria_token_decoded = token_service.decode_jwt(
            decoded["payload"].get("alastriaToken")
        )
        if not token_service.verify_jwt(
            decoded["payload"].get("alastriaToken"), settings.ISSUER_PUBLIC_KEY
        ):
            return HttpResponseBadRequest("Invalid Alastria Token.")
        public_key: str = decoded.get("payload").get("pku")
        request_session_id: int = alastria_token_decoded.get("payload").get(
            "request_session_id"
        )
        if token_service.verify_jwt(token, public_key):
            address: str = AlastriaAuthService.public_key_to_address(public_key)
            did: str = AlastriaAuthService.get_did(address)
            rs: RequestSession = RequestSession.objects.filter(
                pk=request_session_id
            ).first()
            network_data: NetworkDTO = NetworkClient(
                service_host=settings.NETWORK_SERVICE_HOST
            ).get_network_by_name(NetworksNames.AlastriaDefaultName)
            format_did: str = network_data.did_prefix + did[2:]
            rs.requested_by = format_did
            rs.save()
            return AlastriaAuthService.validate_request_session(
                rs,
                format_did,
                AlastriaAuthService.validate_did(did, Response({"response": "OK"})),
            )

        else:
            return HttpResponseBadRequest("Invalid Token.")
